import { EventEmitter } from 'events';
import { Server } from './';
const TOPIC_PART_LIMIT = 16;
function matchRule(rule, topic) {
  if (rule === topic) {
    return true;
  }
  const ruleParts = rule.split('/');
  const topicParts = topic.split('/');
  for (let i = 0; i < ruleParts.length; i++) {
    if (ruleParts[i] === '#') {
      return true;
    }
    if (ruleParts[i] !== '+' && ruleParts[i] !== topicParts[i]) {
      return false;
    }
  }
  return ruleParts.length === topicParts.length;
}
const verifyTopic = topic => {
  if (/^[^#]*(?:\/?#|\/?[^#]*)?$/.test(topic)) {
    return topic.split('/').length <= TOPIC_PART_LIMIT;
  } else {
    return false;
  }
};
export class SimpleMQBroker extends EventEmitter {
  constructor() {
    let options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    super();
    this.server = new Server();
    this.authenticate = options.authenticate;
    this.retryInterval = options.retryInterval ?? 3000;
    this.keepalive = options.keepalive ?? 60;
    this.clients = {};
    this.server.on('connection', this.handleClient.bind(this));
    this.server.on('error', this.emit.bind(this, 'error'));
    this.server.on('close', this.emit.bind(this, 'close'));
    this.server.on('listening', this.emit.bind(this, 'listening'));
    this.sessions = {};
    this.nextId = 1000;
  }
  start(port) {
    this.server.listen({
      port: port ?? 1883,
      host: '0.0.0.0'
    });
  }
  stop() {
    this.server.close();
  }
  handleClient(client) {
    let id;
    let session;
    client.on('close', () => {
      if (id && session) {
        delete this.clients[id];
        session.alive = false;
      }
    });
    client.on('disconnect', () => {
      client.end();
      this.emit('clientDisconnected', id);
    });
    client.on('connect', async packet => {
      var _this$sessions$packet, _this$authenticate;
      if ((_this$sessions$packet = this.sessions[packet.clientId]) !== null && _this$sessions$packet !== void 0 && _this$sessions$packet.alive) {
        client.connack({
          returnCode: 2
        });
        return;
      }
      const version = packet.protocolVersion ?? 5;
      client.setProtocolVersion(version);
      const pass = (await ((_this$authenticate = this.authenticate) === null || _this$authenticate === void 0 ? void 0 : _this$authenticate.call(this, packet.username, packet.password))) ?? true;
      const sessionPresent = packet.clean ? false : !!this.sessions[packet.clientId];
      const keepalive = packet.keepalive ?? this.keepalive;
      if (pass) {
        client.connack({
          returnCode: 0,
          reasonCode: 0,
          sessionPresent,
          attributes: {
            serverKeepAlive: keepalive
          }
        });
        client.setKeepAlive(keepalive);
      } else {
        client.connack({
          returnCode: 5,
          reasonCode: 135
        });
      }
      id = `${packet.username ?? 'anonymous'}-${packet.clientId}`;
      this.clients[id] = client;
      session = this.sessions[id] ??= {
        subs: {},
        alive: true,
        pending: {},
        tasks: {}
      };
      session.alive = true;
      if (packet.clean) {
        Object.values(session.tasks).forEach(clearInterval);
        session.tasks = {};
        session.subs = {};
        session.pending = {};
      }
      this.emit('clientConnected', id);
    });
    client.on('publish', async packet => {
      if (!session) return;
      switch (packet.qos) {
        case 0:
          this.processMessage(packet);
          break;
        case 1:
          client.puback({
            messageId: packet.messageId
          });
          this.processMessage(packet);
          break;
        case 2:
          session.pending[packet.messageId] = {
            topic: packet.topic,
            payload: packet.payload
          };
          client.pubrec({
            messageId: packet.messageId
          });
          break;
      }
    });
    client.on('pubrel', packet => {
      var _session, _session2;
      const {
        messageId
      } = packet;
      if (!((_session = session) !== null && _session !== void 0 && _session.pending[messageId])) return;
      const message = session.pending[messageId];
      (_session2 = session) === null || _session2 === void 0 ? true : delete _session2.pending[messageId];
      client.pubcomp({
        messageId
      });
      this.processMessage({
        ...message,
        messageId
      });
    });
    client.on('subscribe', packet => {
      if (!session) return;
      const granted = [];
      for (const sub of packet.subscriptions) {
        if (!verifyTopic(sub.topic)) {
          granted.push(128);
        } else {
          session.subs[sub.topic] = sub.qos;
          granted.push(sub.qos);
        }
      }
      client.suback({
        messageId: packet.messageId,
        granted
      });
    });
    client.on('unsubscribe', packet => {
      if (!session) return;
      if (Object.keys(session.subs).length === 0) {
        client.unsuback({
          messageId: packet.messageId,
          reasonCode: 17
        });
        return;
      }
      const granted = [];
      for (const topic of packet.unsubscriptions) {
        if (!session.subs[topic]) {
          granted.push(128);
        } else {
          delete session.subs[topic];
          granted.push(0);
        }
      }
      client.unsuback({
        messageId: packet.messageId,
        reasonCode: 0,
        granted
      });
    });
    client.on('pingreq', () => {
      client.pingresp();
    });
    client.on('puback', packet => {
      var _session3;
      const messageId = packet.messageId ? Number(packet.messageId) : -1;
      if (((_session3 = session) === null || _session3 === void 0 ? void 0 : _session3.tasks[messageId]) !== undefined) {
        clearInterval(session.tasks[messageId]);
        delete session.tasks[messageId];
        this.emit('messageSend', id, messageId);
      }
    });
    client.on('pubrec', packet => {
      var _session4;
      const messageId = packet.messageId ? Number(packet.messageId) : -1;
      if (((_session4 = session) === null || _session4 === void 0 ? void 0 : _session4.tasks[messageId]) !== undefined) {
        client === null || client === void 0 ? void 0 : client.pubrel({
          messageId
        });
      }
    });
    client.on('pubcomp', packet => {
      var _session5;
      const messageId = packet.messageId ? Number(packet.messageId) : -1;
      if (((_session5 = session) === null || _session5 === void 0 ? void 0 : _session5.tasks[messageId]) !== undefined) {
        clearInterval(session.tasks[messageId]);
        delete session.tasks[messageId];
        this.emit('messageSend', id, messageId);
      }
    });
  }
  processMessage(packet) {
    this.emit('message', packet.topic, packet.payload);
    this.publish(packet.topic, packet.payload, packet.messageId);
  }
  publish(topic, payload, messageId) {
    const id = messageId ?? this.nextId++;
    for (const cId in this.sessions) {
      for (const rule in this.sessions[cId].subs) {
        if (matchRule(rule, topic)) {
          const qos = this.sessions[cId].subs[rule] ?? 0;
          this.createPubTask(cId, id, topic, payload, qos);
          break;
        }
      }
    }
    return id;
  }
  createPubTask(cId, messageId, topic, payload, qos) {
    let dup = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : false;
    const session = this.sessions[cId];
    if (!(session !== null && session !== void 0 && session.alive)) return;
    const pubPacket = {
      messageId,
      topic,
      payload,
      qos,
      dup
    };
    const client = this.clients[cId];
    if (qos > 0) {
      session.tasks[messageId] = setInterval(() => {
        client === null || client === void 0 ? void 0 : client.publish(pubPacket);
      }, this.retryInterval);
    }
    client === null || client === void 0 ? void 0 : client.publish(pubPacket);
  }
}
export default SimpleMQBroker;
//# sourceMappingURL=simple-mq.js.map